==> klayout-0.10.tar.gz: klayout-0.10/src/layLayerTreeModel.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layLayerTreeModel.h"
#include "layLayoutView.h"
#include "layBitmapsToImage.h"
#include "tlLog.h"

#include <QtGui/QTreeView>
#include <QtCore/QModelIndex>
#include <QtCore/QString>

#include <string.h>

#include <algorithm>

namespace lay {

// --------------------------------------------------------------------
//  LayerTreeModel implementation

LayerTreeModel::LayerTreeModel (QWidget *parent, lay::LayoutView *view)
  : QAbstractItemModel (parent), 
    mp_view (view), m_phase ((unsigned int) -1)
{
  // .. nothing yet ..
}

LayerTreeModel::~LayerTreeModel ()
{
  // .. nothing yet ..
}

void
LayerTreeModel::set_phase (unsigned int ph)
{
  m_phase = ph;
}

void 
LayerTreeModel::set_font (const QFont &font)
{
  m_font = font;
  signal_data_changed ();
}

void 
LayerTreeModel::set_text_color (QColor color)
{
  m_text_color = color;
  signal_data_changed ();
}

void 
LayerTreeModel::set_background_color (QColor background)
{
  m_background_color = background;
  signal_data_changed ();
}

Qt::ItemFlags 
LayerTreeModel::flags (const QModelIndex &index) const
{
  return QAbstractItemModel::flags (index);
}

int 
LayerTreeModel::columnCount (const QModelIndex &) const 
{
  return 2;
}

QVariant 
LayerTreeModel::headerData (int /*section*/, Qt::Orientation /*orientation*/, int /*role*/) const
{
  return QVariant ();
}

int 
LayerTreeModel::rowCount (const QModelIndex &parent) const 
{
  try {
    mp_view->check_updated ();
    if (parent.isValid ()) {
      lay::LayerPropertiesConstIterator iter (iterator (parent));
      return iter->end_children () - iter->begin_children ();
    } else {
      int n = int (mp_view->get_properties ().end_const () - mp_view->get_properties ().begin_const ());
      return n;
    }
  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return 0;
  }
}

QModelIndex 
LayerTreeModel::index (int row, int column, const QModelIndex &parent) const 
{
  try {
    mp_view->check_updated ();
    if (parent.isValid ()) {
      lay::LayerPropertiesConstIterator iter (iterator (parent));
      iter.first_child ();
      iter.next_sibling (row);
      return createIndex (row, column, int (iter.uint ()));
    } else {
      lay::LayerPropertiesConstIterator iter (mp_view->begin_layers ());
      iter.next_sibling (row);
      return createIndex (row, column, int (iter.uint ()));
    }
  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return QModelIndex ();
  }
}

void 
LayerTreeModel::signal_data_changed () 
{
  emit dataChanged (upperLeft (), bottomRight ());
}

void 
LayerTreeModel::signal_begin_layer_changed () 
{
  emit layoutAboutToBeChanged ();
}

void 
LayerTreeModel::signal_layer_changed () 
{
  emit layoutChanged ();
}

QModelIndex
LayerTreeModel::upperLeft () const
{
  try {
    mp_view->check_updated ();
    lay::LayerPropertiesConstIterator iter (mp_view->begin_layers ());
    iter.next_sibling (0);
    return createIndex (0, 0, int (iter.uint ()));
  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return QModelIndex ();
  }
}

QModelIndex
LayerTreeModel::bottomRight () const
{
  try {
    mp_view->check_updated ();
    lay::LayerPropertiesConstIterator iter (mp_view->begin_layers ());
    int n = int (mp_view->get_properties ().end_const () - mp_view->get_properties ().begin_const ()) - 1;
    iter.next_sibling (n);
    return createIndex (n, 1, int (iter.uint ()));
  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return QModelIndex ();
  }
}

QModelIndex 
LayerTreeModel::parent (const QModelIndex &index) const 
{
  try {
    mp_view->check_updated ();
    lay::LayerPropertiesConstIterator iter (iterator (index));
    iter.up ();
    if (iter.is_null ()) {
      return QModelIndex ();
    } else {
      //  It is important that the column index of the parent is 0.
      //  Otherwise the tree view will not behave as expected.
      return createIndex (iter.child_index (), 0, int (iter.uint ()));
    }
  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return QModelIndex ();
  }
}

/**
 *  @brief A helper function to create an image from a single bitmap
 */
static void
single_bitmap_to_image (const lay::ViewOp &view_op, lay::Bitmap &bitmap,
                        QImage *pimage, const lay::DitherPattern &dither_pattern, 
                        unsigned int width, unsigned int height)
{
  std::vector <lay::ViewOp> view_ops;
  view_ops.push_back (view_op);

  std::vector <lay::Bitmap *> pbitmaps;
  pbitmaps.push_back (&bitmap);

  lay::bitmaps_to_image (view_ops, pbitmaps, dither_pattern, pimage, width, height, 0);
}

QVariant 
LayerTreeModel::data (const QModelIndex &index, int role) const 
{
  try {

    mp_view->check_updated ();

    lay::LayerPropertiesConstIterator iter (iterator (index));
    if (role == Qt::DisplayRole || role == Qt::EditRole) {

      if (index.column () == 1) {
        return QVariant (QString (iter->display_string (true /*real*/).c_str ()));
      }

    } else if (role == Qt::DecorationRole) {

      if (index.column () == 0) {

        bool animate_visible = true;
        unsigned int di_off = 0;
        if (iter->animation (true)) {
          if (iter->animation (true) == 1) {
            // scrolling 
            di_off = ((m_phase % 32) << 16);
          } else if (iter->animation (true) == 2) {
            // blinking
            animate_visible = ((m_phase & 1) == 0);
          } else {
            // inversly blinking
            animate_visible = ((m_phase & 1) != 0);
          }
        }

        unsigned int w = 32;
        unsigned int h = 16;
          
        if (iter->visible (true) && animate_visible) {

          lay::Bitmap fill (w, h);
          for (unsigned int i = 1; i < h - 1; ++i) {
            fill.scanline (i) [0] = 0xffffffff;
          }

          int p0 = iter->width (true) / 2;
          int p1 = (iter->width (true) - 1) / 2;
          if (p0 < 0) {
            p0 = 0;
          } else if (p0 > 7) {
            p0 = 7;
          }
          if (p1 < 0) {
            p1 = 0;
          } else if (p1 > 7) {
            p1 = 7;
          }

          lay::Bitmap frame (w, h);
          frame.scanline (p0) [0] = 0xffffffff;
          for (unsigned int i = p0 + 1; i < h - 1; ++i) {
            frame.scanline (i) [0] = (0x80000000 >> p0) | (1 << p1);
          }
          frame.scanline (15 - p1) [0] = 0xffffffff;

          lay::Bitmap text (w, h);
          lay::Bitmap vertex (w, h);
          vertex.scanline (7) [0] = 0x00010000;

          lay::ViewOp::Mode mode = lay::ViewOp::Copy;

          QImage image (w, h, 32);
          image.fill (m_background_color.rgb ());

          lay::color_t def_color   = 0x808080;
          lay::color_t fill_color  = iter->has_fill_color (true)  ? iter->eff_fill_color (true)  : def_color;
          lay::color_t frame_color = iter->has_frame_color (true) ? iter->eff_frame_color (true) : def_color;

          //  create fill 
          single_bitmap_to_image (lay::ViewOp (fill_color, mode, iter->eff_dither_pattern (true) | di_off), fill, &image, mp_view->dither_pattern (), w, h);
          //  create frame 
          if (iter->width (true) == 0) {
            single_bitmap_to_image (lay::ViewOp (frame_color, mode, 2 /*dotted*/), frame, &image, mp_view->dither_pattern (), w, h);
          } else {
            single_bitmap_to_image (lay::ViewOp (frame_color, mode, 0, lay::ViewOp::Rect, iter->width (true)), frame, &image, mp_view->dither_pattern (), w, h);
          }
          //  create text 
          single_bitmap_to_image (lay::ViewOp (frame_color, mode, 0), text, &image, mp_view->dither_pattern (), w, h);
          //  create vertex 
          single_bitmap_to_image (lay::ViewOp (frame_color, mode, 0, lay::ViewOp::Cross, iter->marked (true) ? 9/*mark size*/ : 0), vertex, &image, mp_view->dither_pattern (), w, h);

          return QVariant (QIcon (QPixmap (image)));

        } else {
          return QVariant (QIcon ());
        }

      }

    } else if (role == Qt::TextColorRole || role == Qt::FontRole) {

      if (index.column () == 1) {

        QColor c1 = m_text_color;
        QColor cb = m_background_color;
        QColor c0 = QColor ((c1.red () + cb.red ()) / 2, (c1.green () + cb.green ()) / 2, (c1.blue () + cb.blue ()) / 2);

        //  set the color to one with less contrast if there is nothing on this layer here
        bool empty = iter->bbox ().empty ();
        if (role == Qt::FontRole) {
          QFont f = m_font;
          f.setBold (! empty);
          return QVariant (f);
        } else {
          return QVariant (empty ? c0 : c1);
        }

      }

    } 

  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
  }

  return QVariant ();
}

lay::LayerPropertiesConstIterator 
LayerTreeModel::iterator (const QModelIndex &index) const
{
  return lay::LayerPropertiesConstIterator (mp_view->get_properties (), size_t (index.internalId ()));
}

QModelIndex 
LayerTreeModel::index (lay::LayerPropertiesConstIterator iter) const
{
  try {

    std::vector<int> rows;
    
    while (! iter.is_null ()) {
      rows.push_back (int (iter.child_index ()));
      iter = iter.parent ();
    }

    QModelIndex idx;

    for (std::vector<int>::reverse_iterator r = rows.rbegin (); r != rows.rend (); ++r) {
      idx = index (*r, 0, idx);
    }

    return idx;

  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return QModelIndex ();
  }
}

} // namespace lay

==> klayout-0.15.tar.gz: klayout-0.15/src/layLayerTreeModel.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layLayerTreeModel.h"
#include "layLayoutView.h"
#include "layBitmapsToImage.h"
#include "tlLog.h"

#include <QtGui/QTreeView>
#include <QtCore/QModelIndex>
#include <QtCore/QString>

#include <string.h>

#include <algorithm>

namespace lay {

// --------------------------------------------------------------------
//  LayerTreeModel implementation

LayerTreeModel::LayerTreeModel (QWidget *parent, lay::LayoutView *view)
  : QAbstractItemModel (parent), 
    mp_view (view), m_phase ((unsigned int) -1)
{
  // .. nothing yet ..
}

LayerTreeModel::~LayerTreeModel ()
{
  // .. nothing yet ..
}

void
LayerTreeModel::set_phase (unsigned int ph)
{
  m_phase = ph;
}

void 
LayerTreeModel::set_font (const QFont &font)
{
  m_font = font;
  signal_data_changed ();
}

void 
LayerTreeModel::set_text_color (QColor color)
{
  m_text_color = color;
  signal_data_changed ();
}

void 
LayerTreeModel::set_background_color (QColor background)
{
  m_background_color = background;
  signal_data_changed ();
}

Qt::ItemFlags 
LayerTreeModel::flags (const QModelIndex &index) const
{
  return QAbstractItemModel::flags (index);
}

int 
LayerTreeModel::columnCount (const QModelIndex &) const 
{
  return 2;
}

QVariant 
LayerTreeModel::headerData (int /*section*/, Qt::Orientation /*orientation*/, int /*role*/) const
{
  return QVariant ();
}

int 
LayerTreeModel::rowCount (const QModelIndex &parent) const 
{
  if (mp_view->layer_model_updated ()) {

    if (parent.isValid ()) {
      lay::LayerPropertiesConstIterator iter (iterator (parent));
      return iter->end_children () - iter->begin_children ();
    } else {
      int n = int (mp_view->get_properties ().end_const () - mp_view->get_properties ().begin_const ());
      return n;
    }

  } else {
    return 0;
  }
}

QModelIndex 
LayerTreeModel::index (int row, int column, const QModelIndex &parent) const 
{
  if (mp_view->layer_model_updated ()) {
    if (parent.isValid ()) {
      lay::LayerPropertiesConstIterator iter (iterator (parent));
      iter.down_first_child ();
      iter.next_sibling (row);
      return createIndex (row, column, int (iter.uint ()));
    } else {
      lay::LayerPropertiesConstIterator iter (mp_view->begin_layers ());
      iter.next_sibling (row);
      return createIndex (row, column, int (iter.uint ()));
    }
  } else {
    return QModelIndex ();
  }
}

void 
LayerTreeModel::signal_data_changed () 
{
  emit dataChanged (upperLeft (), bottomRight ());
}

void 
LayerTreeModel::signal_begin_layer_changed () 
{
  emit layoutAboutToBeChanged ();
}

void 
LayerTreeModel::signal_layer_changed () 
{
  emit layoutChanged ();
}

QModelIndex
LayerTreeModel::upperLeft () const
{
  if (mp_view->layer_model_updated ()) {
    lay::LayerPropertiesConstIterator iter (mp_view->begin_layers ());
    iter.next_sibling (0);
    return createIndex (0, 0, int (iter.uint ()));
  } else {
    return QModelIndex ();
  }
}

QModelIndex
LayerTreeModel::bottomRight () const
{
  if (mp_view->layer_model_updated ()) {

    //  navigate to the last top-level item
    lay::LayerPropertiesConstIterator iter (mp_view->begin_layers ());
    int n = int (mp_view->get_properties ().end_const () - mp_view->get_properties ().begin_const ()) - 1;
    iter.next_sibling (n);

    //  navigate to the last child 
    QModelIndex p = createIndex (n, 1, int (iter.uint ()));
    int nr = 0;
    while (p.isValid () && (nr = rowCount (p)) > 0) {
      p = index (nr - 1, 0, p);
    }
    return p;

  } else {
    return QModelIndex ();
  }
}

QModelIndex 
LayerTreeModel::parent (const QModelIndex &index) const 
{
  if (mp_view->layer_model_updated ()) {
    lay::LayerPropertiesConstIterator iter (iterator (index));
    iter.up ();
    if (iter.is_null ()) {
      return QModelIndex ();
    } else {
      //  It is important that the column index of the parent is 0.
      //  Otherwise the tree view will not behave as expected.
      return createIndex (iter.child_index (), 0, int (iter.uint ()));
    }
  } else {
    return QModelIndex ();
  }
}

/**
 *  @brief A helper function to create an image from a single bitmap
 */
static void
single_bitmap_to_image (const lay::ViewOp &view_op, lay::Bitmap &bitmap,
                        QImage *pimage, const lay::DitherPattern &dither_pattern, 
                        unsigned int width, unsigned int height)
{
  std::vector <lay::ViewOp> view_ops;
  view_ops.push_back (view_op);

  std::vector <lay::Bitmap *> pbitmaps;
  pbitmaps.push_back (&bitmap);

  lay::bitmaps_to_image (view_ops, pbitmaps, dither_pattern, pimage, width, height, false, 0);
}

QVariant 
LayerTreeModel::data (const QModelIndex &index, int role) const 
{
  if (mp_view->layer_model_updated ()) {

    lay::LayerPropertiesConstIterator iter (iterator (index));
    if (role == Qt::DisplayRole || role == Qt::EditRole) {

      if (index.column () == 1) {
        return QVariant (QString (iter->display_string (mp_view, true /*real*/).c_str ()));
      }

    } else if (role == Qt::DecorationRole) {

      if (index.column () == 0) {

        bool animate_visible = true;
        unsigned int di_off = 0;
        if (iter->animation (true)) {
          if (iter->animation (true) == 1) {
            // scrolling 
            di_off = ((m_phase % 32) << 16);
          } else if (iter->animation (true) == 2) {
            // blinking
            animate_visible = ((m_phase & 1) == 0);
          } else {
            // inversly blinking
            animate_visible = ((m_phase & 1) != 0);
          }
        }

        unsigned int w = 32;
        unsigned int h = 16;
          
        if (iter->visible (true) && animate_visible) {

          lay::Bitmap fill (w, h);
          for (unsigned int i = 1; i < h - 1; ++i) {
            fill.scanline (i) [0] = 0xffffffff;
          }

          int p0 = iter->width (true) / 2;
          int p1 = (iter->width (true) - 1) / 2;
          if (p0 < 0) {
            p0 = 0;
          } else if (p0 > 7) {
            p0 = 7;
          }
          if (p1 < 0) {
            p1 = 0;
          } else if (p1 > 7) {
            p1 = 7;
          }

          lay::Bitmap frame (w, h);
          frame.scanline (p0) [0] = 0xffffffff;
          for (unsigned int i = p0 + 1; i < h - 1; ++i) {
            frame.scanline (i) [0] = (0x80000000 >> p0) | (1 << p1);
          }
          frame.scanline (15 - p1) [0] = 0xffffffff;

          lay::Bitmap text (w, h);
          lay::Bitmap vertex (w, h);
          vertex.scanline (7) [0] = 0x00010000;

          lay::ViewOp::Mode mode = lay::ViewOp::Copy;

          QImage image (w, h, 32);
          image.fill (m_background_color.rgb ());

          lay::color_t def_color   = 0x808080;
          lay::color_t fill_color  = iter->has_fill_color (true)  ? iter->eff_fill_color (true)  : def_color;
          lay::color_t frame_color = iter->has_frame_color (true) ? iter->eff_frame_color (true) : def_color;

          //  create fill 
          single_bitmap_to_image (lay::ViewOp (fill_color, mode, iter->eff_dither_pattern (true) | di_off), fill, &image, mp_view->dither_pattern (), w, h);
          //  create frame 
          if (iter->width (true) == 0) {
            single_bitmap_to_image (lay::ViewOp (frame_color, mode, 2 /*dotted*/), frame, &image, mp_view->dither_pattern (), w, h);
          } else {
            single_bitmap_to_image (lay::ViewOp (frame_color, mode, 0, lay::ViewOp::Rect, iter->width (true)), frame, &image, mp_view->dither_pattern (), w, h);
          }
          //  create text 
          single_bitmap_to_image (lay::ViewOp (frame_color, mode, 0), text, &image, mp_view->dither_pattern (), w, h);
          //  create vertex 
          single_bitmap_to_image (lay::ViewOp (frame_color, mode, 0, lay::ViewOp::Cross, iter->marked (true) ? 9/*mark size*/ : 0), vertex, &image, mp_view->dither_pattern (), w, h);

          return QVariant (QIcon (QPixmap (image)));

        } else {
          return QVariant (QIcon ());
        }

      }

    } else if (role == Qt::TextColorRole || role == Qt::FontRole) {

      if (index.column () == 1) {

        QColor c1 = m_text_color;
        QColor cb = m_background_color;
        QColor c0 = QColor ((c1.red () + cb.red ()) / 2, (c1.green () + cb.green ()) / 2, (c1.blue () + cb.blue ()) / 2);

        //  set the color to one with less contrast if there is nothing on this layer here
        bool empty = iter->bbox ().empty ();
        if (role == Qt::FontRole) {
          QFont f = m_font;
          f.setBold (! empty);
          return QVariant (f);
        } else {
          return QVariant (empty ? c0 : c1);
        }

      }

    } 

  }

  return QVariant ();
}

lay::LayerPropertiesConstIterator 
LayerTreeModel::iterator (const QModelIndex &index) const
{
  return lay::LayerPropertiesConstIterator (mp_view->get_properties (), size_t (index.internalId ()));
}

QModelIndex 
LayerTreeModel::index (lay::LayerPropertiesConstIterator iter, int column) const
{
  try {

    std::vector<int> rows;
    
    while (! iter.is_null ()) {
      rows.push_back (int (iter.child_index ()));
      iter = iter.parent ();
    }

    QModelIndex idx;

    for (std::vector<int>::reverse_iterator r = rows.rbegin (); r != rows.rend (); ++r) {
      idx = index (*r, column, idx);
    }

    return idx;

  } catch (tl::Exception &ex) {
    tl::warn << ex.msg ();
    return QModelIndex ();
  }
}

} // namespace lay
